import copy
import json
import logging
import os
from os import path
from unittest.mock import Mock
from fylesdk import FyleSDK
from fyle_db_connector import FyleExtractConnector, FyleLoadConnector

logger = logging.getLogger(__name__)


def dict_factory(cursor, row):
    """
    Sqlite dictionary row factory
    """
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


def get_mock_fyle_dict(filename):
    """
    get the mock_fyle dictionary
    :param filename: filename
    :return: mock_fyle_dict
    """
    basepath = path.dirname(__file__)
    filepath = path.join(basepath, filename)
    mock_fyle_json = open(filepath, 'r').read()
    mock_fyle_dict = json.loads(mock_fyle_json)
    return mock_fyle_dict


def get_mock_fyle_from_file(filename):
    """
    get mock_fyle dictionary from the mock_file provided
    :param filename: filename
    :return: mock_fyle
    """
    mock_fyle_dict = get_mock_fyle_dict(filename)
    mock_fyle = Mock()
    mock_fyle.Expenses.get_all.return_value = mock_fyle_dict['expenses_get_all']
    mock_fyle.Settlements.get_all.return_value = mock_fyle_dict['settlements_get_all']
    mock_fyle.Reimbursements.get_all.return_value = mock_fyle_dict['reimbursements_get_all']
    mock_fyle.Employees.get_all.return_value = mock_fyle_dict['employees_get_all']
    mock_fyle.Advances.get_all.return_value = mock_fyle_dict['advances_get_all']
    mock_fyle.AdvanceRequests.get_all.return_value = mock_fyle_dict['advance_requests_get_all']
    mock_fyle.Projects.get.return_value = mock_fyle_dict['projects_get']
    mock_fyle.CostCenters.get.return_value = mock_fyle_dict['cost_centers_get']
    mock_fyle.Categories.get.return_value = mock_fyle_dict['categories_get']
    mock_fyle.CorporateCreditCardExpenses.get_all.return_value = mock_fyle_dict[
        'corporate_cards_get_all']
    mock_fyle.Exports.post_batch_lineitems.save.return_value = copy.deepcopy(
        mock_fyle_dict['post_batch_lineitems_response'])
    mock_fyle.Exports.post_batch.return_value = copy.deepcopy(mock_fyle_dict['post_batch_sdk_response'])
    mock_fyle.Exports.post_batch.save.return_value = copy.deepcopy(mock_fyle_dict['post_batch_response'])
    mock_fyle.Exports.post_batch_lineitems.return_value = copy.deepcopy(
        mock_fyle_dict['post_batch_lineitems_sdk_response'])

    return mock_fyle


def get_mock_fyle():
    """
    get mock_fyle containing mock_data
    :return: mock_fyle
    """
    return get_mock_fyle_from_file('mock_fyle.json')


def get_mock_fyle_empty():
    """
    get empty mock_fyle
    :return: mock_fyle
    """
    return get_mock_fyle_from_file('mock_fyle_empty.json')


def dict_compare_keys(d1, d2, key_path=''):
    """
    Compare two dicts recursively and see if dict1 has any keys that dict2 does not
    :returns: list of key paths
    """
    res = []
    if not d1:
        return res
    if not isinstance(d1, dict):
        return res
    for k in d1:
        if k not in d2:
            missing_key_path = f'{key_path}->{k}'
            res.append(missing_key_path)
        else:
            if isinstance(d1[k], dict):
                key_path1 = f'{key_path}->{k}'
                res1 = dict_compare_keys(d1[k], d2[k], key_path1)
                res = res + res1
            elif isinstance(d1[k], list):
                key_path1 = f'{key_path}->{k}[0]'
                dv1 = d1[k][0] if len(d1[k]) > 0 else None
                dv2 = d2[k][0] if len(d2[k]) > 0 else None
                res1 = dict_compare_keys(dv1, dv2, key_path1)
                res = res + res1
    return res


def dbconn_table_num_rows(dbconn, tablename):
    """
    Helper function to calculate number of rows
    """
    query = f'select count(*) from {tablename}'
    return dbconn.cursor().execute(query).fetchone()[0]


def dbconn_table_row_dict(dbconn, tablename):
    """
    get dictionary first row of the table
    :param dbconn: database connection
    :param tablename: name of the table
    :return: dict containing the first row of the table
    """
    query = f'select * from {tablename} limit 1'
    row = dbconn.cursor().execute(query).fetchone()
    return dict(row)


def dbconn_get_load_object_by_id(dbconn, tablename, object_id):
    """
    Get load object by Id
    """
    dbconn.row_factory = dict_factory
    query = f"select * from {tablename} where id = '{object_id}'"
    row = dbconn.cursor().execute(query).fetchall()
    return row


def dbconn_get_load_object_by_batch_id(dbconn, tablename, batch_id):
    """
    Get load object by batch_id
    """
    dbconn.row_factory = dict_factory
    query = f"select * from {tablename} where batch_id = '{batch_id}'"
    row = dbconn.cursor().execute(query).fetchall()
    return row


def fyle_connect():
    """
    establish connection with fyle
    :return: fyle_connection
    """
    # Initializing FyleSDK
    fyle_connection = FyleSDK(
        base_url=os.environ.get('FYLE_TPA_BASE_URL'),
        client_id=os.environ.get('FYLE_TPA_CLIENT_ID'),
        client_secret=os.environ.get('FYLE_TPA_CLIENT_SECRET'),
        refresh_token=os.environ.get('FYLE_TPA_REFRESH_TOKEN')
    )
    return fyle_connection


def fec(dbconn):
    """

    :param dbconn: database connection
    :return: fyle_extract
    """
    fyle_extract = FyleExtractConnector(fyle_sdk_connection=fyle_connect(), dbconn=dbconn)
    return fyle_extract


def flc(dbconn):
    """
    Fyle Connector connection
    :param dbconn:
    :return:
    """
    fyle_load = FyleLoadConnector(fyle_sdk_connection=fyle_connect(), dbconn=dbconn)
    return fyle_load


def execute_sqlfile(dbconn, file_path):
    """
    execute sql file
    :param dbconn: database connection
    :param file_path: path of the sql file
    :return: sql file
    """
    with open(file_path, 'r') as tf:
        tsql = tf.read()
        sql = tsql
        logger.debug('executing sql %s', sql)
        dbconn.cursor().executescript(sql)



"""
FyleExtractConnector(): Connection between Fyle and Database
"""

import logging
from os import path
import sqlite3
from typing import List
import pandas as pd


class FyleExtractConnector:
    """
    - Extract Data from Fyle and load to Database
    """

    def __init__(self, fyle_sdk_connection, dbconn):
        self.__dbconn = dbconn
        self.__connection = fyle_sdk_connection
        self.__dbconn.row_factory = sqlite3.Row

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info('Fyle connection established')

    def create_tables(self):
        """
        Creates DB tables
        :return: None
        """
        basepath = path.dirname(__file__)
        ddl_path = path.join(basepath, 'extract_ddl.sql')
        ddl_sql = open(ddl_path, 'r').read()
        self.__dbconn.executescript(ddl_sql)

    def extract_settlements(self, updated_at: List[str] = None, exported: bool = None) -> List[str]:
        """
        Extract settlements from Fyle
        :param updated_at: Date string in yyyy-MM-ddTHH:mm:ss.SSSZ format along with operator in RHS colon pattern.
        :param exported: True for exported settlements and False for unexported settlements
        :return: List of settlement ids
        """

        self.logger.info('Extracting settlements from Fyle.')

        settlements = self.__connection.Settlements.get_all(updated_at=updated_at, exported=exported)

        df_settlements = pd.DataFrame(settlements)
        self.logger.info('%s settlements extracted.', str(len(df_settlements.index)))

        if settlements:
            df_settlements = df_settlements[[
                'id', 'created_at', 'updated_at', 'opening_date', 'closing_date',
                'employee_id', 'employee_email', 'employee_code', 'creator_employee_id',
                'creator_employee_email', 'creator_employee_code', 'org_id', 'org_name',
                'exported'
            ]]

            df_settlements.to_sql('fyle_extract_settlements', self.__dbconn, if_exists='append', index=False)
            return df_settlements['id'].to_list()

        return []

    def extract_employees(self, extract_custom_fields: bool = True) -> List[str]:
        """
        Extract employees from Fyle
        :return: List of employee ids
        """

        self.logger.info('Extracting employees from Fyle.')
        employees = self.__connection.Employees.get_all()

        self.logger.info('%s employees extracted.', str(len(employees)))

        if employees:
            df_employees = pd.DataFrame(employees)

            df_employees = df_employees[[
                'id', 'created_at', 'updated_at', 'employee_email', 'employee_code',
                'full_name', 'joining_date', 'location', 'level_id', 'level',
                'business_unit', 'department_id', 'department', 'sub_department',
                'approver1_email', 'approver2_email', 'approver3_email', 'title',
                'branch_ifsc', 'branch_account', 'mobile', 'delegatee_email',
                'default_cost_center_name', 'disabled', 'org_id', 'org_name'
            ]]

            df_employees.to_sql('fyle_extract_employees', self.__dbconn, if_exists='append', index=False)

            custom_fields = []
            for e in employees:
                if 'custom_fields' in e and e['custom_fields']:
                    for cp in e['custom_fields']:
                        custom_fields.append({
                            'employee_id': e['id'],
                            'name': cp['custom_field_name'],
                            'value': cp['custom_field_value']
                        })

            if extract_custom_fields:
                if custom_fields:
                    df_custom_fields = pd.DataFrame(custom_fields)
                    df_custom_fields = df_custom_fields.astype('str')
                    df_custom_fields.to_sql('fyle_extract_employee_custom_fields', self.__dbconn,
                                            if_exists='append', index=False)

            return df_employees['id'].to_list()

        return []

    def extract_expenses(self, settlement_ids: List[str] = None, state: List[str] = None,
                         fund_source: List[str] = None, reimbursable: bool = None, updated_at: List[str] = None,
                         exported: bool = None, extract_custom_fields: bool = True, spent_at: List[str] = None, report_ids: List[str] = None) -> List[str]:
        """
        Extract expenses from Fyle
        :param extract_custom_fields: Flag to extract custom fields, true by default
        :param updated_at: Extract expenses in exported_at date range
        :param exported: True for exported expenses and False for unexported expenses
        :param settlement_ids: List of settlement_ids
        :param state: List of expense states
        :param fund_source: List of expense fund_sources
        :param reimbursable: True for reimbursable expenses, False for non reimbursable expenses
        :param spent_at: Extract expenses in spent_at date range
        :param report_ids: List of report_ids
        :return: List of expense ids
        """

        self.logger.info('Extracting expenses from Fyle.')

        expenses = self.__connection.Expenses.get_all(
            settlement_id=settlement_ids,
            state=state,
            updated_at=updated_at,
            fund_source=fund_source,
            exported=exported,
            spent_at=spent_at,
            report_id=report_ids
        )

        if reimbursable is not None:
            expenses = list(filter(lambda expense: expense['reimbursable'], expenses))

        self.logger.info('%s expenses extracted.', str(len(expenses)))

        if expenses:
            df_expenses = pd.DataFrame(expenses)

            df_expenses['approved_by'] = df_expenses['approved_by'].map(lambda expense: expense[0] if expense else None)

            df_expenses = df_expenses[[
                'id', 'employee_id', 'employee_email', 'employee_code', 'spent_at', 'currency',
                'amount', 'foreign_currency', 'foreign_amount', 'purpose', 'project_id', 'project_name',
                'cost_center_id', 'cost_center_name', 'category_id', 'category_code', 'category_name',
                'sub_category', 'settlement_id', 'expense_number', 'claim_number', 'trip_request_id',
                'state', 'report_id', 'fund_source', 'reimbursable', 'created_at', 'updated_at',
                'approved_at', 'settled_at', 'split_group_id', 'split_group_user_amount', 'verified',
                'verified_at', 'reimbursed_at', 'added_to_report_at', 'report_submitted_at', 'vendor',
                'has_attachments', 'billable', 'exported', 'approved_by', 'org_id', 'org_name', 'created_by',
                'from_dt', 'to_dt', 'cost_center_code'
            ]]

            df_expenses.to_sql('fyle_extract_expenses', self.__dbconn, if_exists='append', index=False)

            custom_properties = []
            for e in expenses:
                if 'custom_properties' in e and e['custom_properties']:
                    for cp in e['custom_properties']:
                        custom_properties.append({
                            'expense_id': e['id'],
                            'name': cp['name'],
                            'value': cp['value']
                        })

            if extract_custom_fields:
                if custom_properties:
                    df_custom_properties = pd.DataFrame(custom_properties)
                    df_custom_properties = df_custom_properties.astype('str')
                    df_custom_properties.to_sql('fyle_extract_expense_custom_properties', self.__dbconn,
                                                if_exists='append', index=False)

            return df_expenses['id'].to_list()

        return []

    def extract_corporate_credit_card_expenses(self, settlement_ids: List[str] = None, state: List[str] = None,
                                               reimbursement_state: List[str] = None, transaction_type: str = None,
                                               updated_at: List[str] = None, spent_at: List[str] = None,
                                               exported: bool = None, personal: bool = None) -> List[str]:
        """
        Extract corporate credit card expenses from fyle.
        :param settlement_ids: List of settlement_ids
        :param state: state of corporate credit card expense INITIALIZED / IN_PROGRESS / SETTLED
        :param reimbursement_state: state of reimbursement PENDING / COMPLETE
        :param transaction_type: DEBIT / CREDIT
        :param updated_at: Extract expenses in updated_at date range
        :param spent_at: Extract expenses in spent_at date range
        :param exported: True for exported expenses and False for unexported expenses
        :param personal: True for personal expenses and False for company expenses
        :return:
        """

        self.logger.info('Extracting corporate credit card expenses from Fyle.')

        ccc_expenses = self.__connection.CorporateCreditCardExpenses.get_all(
            settlement_id=settlement_ids,
            state=state,
            reimbursement_state=reimbursement_state,
            transaction_type=transaction_type,
            updated_at=updated_at,
            spent_at=spent_at,
            exported=exported,
            personal=personal
        )

        self.logger.info('%s expenses extracted.', str(len(ccc_expenses)))

        if ccc_expenses:
            df_ccc_expenses = pd.DataFrame(ccc_expenses)

            df_ccc_expenses = df_ccc_expenses[[
                'id', 'employee_email', 'org_user_id', 'org_id', 'org_name', 'created_at', 'updated_at', 'spent_at',
                'amount', 'currency', 'foreign_amount', 'foreign_currency', 'vendor', 'description', 'settlement_id',
                'corporate_credit_card_account_number', 'personal', 'expense_id', 'reimbursement_state', 'state',
                'transaction_type', 'exported', 'exported_at', 'expense_split_group_id'
            ]]

            df_ccc_expenses.to_sql('fyle_extract_corporate_credit_card_expenses', self.__dbconn, if_exists='append',
                                   index=False)

            return df_ccc_expenses['id'].to_list()

        return []

    def extract_attachments(self, expense_ids: List[str]) -> List[str]:
        """
        Extract attachments from Fyle
        :param expense_ids: List of Expense Ids
        :return: List of expense ids for which attachments were downloaded
        """
        attachments = []

        self.logger.info('Extracting attachments from Fyle')

        if expense_ids:
            for expense_id in expense_ids:
                attachment = self.__connection.Expenses.get_attachments(expense_id)
                if attachment['data']:
                    attachment = attachment['data'][0]
                    attachment['expense_id'] = expense_id
                    attachments.append(attachment)

            self.logger.info('%s attachments extracted.', str(len(attachments)))

            if attachments:
                df_attachments = pd.DataFrame(attachments)
                df_attachments.to_sql('fyle_extract_attachments', self.__dbconn, if_exists='append', index=False)
                return df_attachments['expense_id'].to_list()

        self.logger.info('0 attachments extracted.')
        return []

    def extract_categories(self) -> List[str]:
        """
        Extract categories from Fyle
        :return: List of category ids
        """
        self.logger.info('Extracting categories from Fyle.')

        categories = self.__connection.Categories.get()['data']

        self.logger.info('%s categories extracted.', str(len(categories)))

        if categories:
            df_categories = pd.DataFrame(categories)

            df_categories = df_categories[[
                'id', 'name', 'code', 'enabled', 'fyle_category', 'sub_category',
                'created_at', 'updated_at', 'org_id', 'org_name'
            ]]

            df_categories.to_sql('fyle_extract_categories', self.__dbconn, if_exists='append', index=False)
            return df_categories['id'].to_list()

        return []

    def extract_projects(self) -> List[str]:
        """
        Extract projects from Fyle
        :return: List of project ids
        """
        self.logger.info('Extracting projects from Fyle.')

        projects = self.__connection.Projects.get()['data']

        self.logger.info('%s projects extracted.', str(len(projects)))

        if projects:
            df_projects = pd.DataFrame(projects)
            df_projects = df_projects[[
                'id', 'name', 'description', 'active', 'approver1_employee_id',
                'approver1_employee_email', 'approver1_employee_code',
                'approver2_employee_id', 'approver2_employee_email',
                'approver2_employee_code', 'org_id', 'code', 'org_name'
            ]]

            df_projects.to_sql('fyle_extract_projects', self.__dbconn, if_exists='append', index=False)
            return df_projects['id'].to_list()

        return []

    def extract_cost_centers(self) -> List[str]:
        """
        Extract cost centers from Fyle
        :return: List of cost center ids
        """
        self.logger.info('Extracting cost centers from Fyle.')

        cost_centers = self.__connection.CostCenters.get(False)['data']

        self.logger.info('%s cost centers extracted.', str(len(cost_centers)))

        if cost_centers:
            df_cost_centers = pd.DataFrame(cost_centers)
            df_cost_centers = df_cost_centers[['id', 'name', 'description', 'code', 'active', 'org_id', 'org_name']]

            df_cost_centers.to_sql('fyle_extract_cost_centers', self.__dbconn, if_exists='append', index=False)
            return df_cost_centers['id'].to_list()

        return []

    def extract_reimbursements(self, state: List[str] = None, updated_at: List[str] = None,
                               exported: bool = None) -> List[str]:
        """
        Extract reimbursements from Fyle
        :param updated_at: Extract expenses within a date range.
        :param state: List of states
        :param exported: True for exported reimbursements and False for unexeported reimbursements
        :return: List of reimbursement ids
        """
        self.logger.info('Extracting reimbursements from Fyle.')
        reimbursements = self.__connection.Reimbursements.get_all(
            updated_at=updated_at,
            exported=exported
        )

        if state:
            reimbursements = list(filter(
                lambda reimbursement: reimbursement['state'] in state,
                reimbursements
            ))

        self.logger.info('%s reimbursements extracted.', str(len(reimbursements)))

        if reimbursements:
            df_reimbursements = pd.DataFrame(reimbursements)

            df_reimbursements = df_reimbursements[[
                'id', 'created_at', 'updated_at', 'employee_id', 'employee_email',
                'employee_code', 'org_id', 'org_name', 'currency', 'amount', 'state',
                'report_ids', 'unique_id', 'purpose', 'settlement_id', 'exported'
            ]]

            df_reimbursements.to_sql('fyle_extract_reimbursements', self.__dbconn, if_exists='append', index=False)
            return df_reimbursements['id'].to_list()

        return []

    def extract_advances(self, settlement_ids: List[str] = None) -> List[str]:
        """
        Extract advances from Fyle
        :param settlement_ids: List of settlement ids
        :return: List of advance ids
        """
        self.logger.info('Extracting advances from Fyle.')
        advances = self.__connection.Advances.get_all(
            settlement_id=settlement_ids
        )

        self.logger.info('%s advances extracted.', str(len(advances)))

        if advances:
            df_advances = pd.DataFrame(advances)

            df_advances = df_advances[[
                'id', 'created_at', 'updated_at', 'employee_id', 'employee_email',
                'employee_code', 'project_id', 'project_name', 'currency', 'amount',
                'purpose', 'issued_at', 'payment_mode', 'original_currency',
                'original_amount', 'reference', 'settlement_id', 'trip_request_id',
                'advance_number', 'advance_request_id', 'settled_at', 'created_by',
                'exported', 'org_id', 'org_name'
            ]]

            df_advances.to_sql('fyle_extract_advances', self.__dbconn, if_exists='append', index=False)
            return df_advances['id'].to_list()

        return []

    def extract_advance_requests(self, state: List[str] = None, updated_at: List[str] = None,
                                 exported: bool = None) -> List[str]:
        """
        Extract advance requests from Fyle
        :param state:
        :param updated_at:
        :param exported:
        :return: List of advance request ids
        """
        self.logger.info('Extracting advance requests from Fyle.')

        advance_requests = self.__connection.AdvanceRequests.get_all(
            state=state,
            updated_at=updated_at,
            exported=exported
        )

        self.logger.info('%s advance requests extracted.', str(len(advance_requests)))

        if advance_requests:
            df_advance_requests = pd.DataFrame(advance_requests)

            if len(df_advance_requests['custom_field_values'].index):
                advance_request_custom_fields = []

                for row in df_advance_requests.to_dict(orient='records'):
                    custom_field = {}
                    custom_fields = row['custom_field_values']
                    custom_field['advance_request_id'] = row['id']

                    if custom_fields:
                        for field in custom_fields:
                            custom_field[field['name']] = field['value']

                        advance_request_custom_fields.append(custom_field)

                df_advance_requests_custom_fields = pd.DataFrame(advance_request_custom_fields)
                df_advance_requests_custom_fields.to_sql(
                    'fyle_extract_advance_request_custom_fields',
                    self.__dbconn,
                    if_exists='replace',
                    index=False
                )

            df_advance_requests = df_advance_requests[[
                'updated_at', 'created_at', 'approved_at', 'id', 'purpose', 'notes',
                'state', 'currency', 'amount', 'advance_id', 'advance_request_number',
                'trip_request_id', 'project_id', 'source', 'is_sent_back', 'is_pulled_back',
                'org_id', 'org_name', 'employee_email', 'employee_name', 'employee_id',
                'exported'
            ]]

            df_advance_requests.to_sql('fyle_extract_advance_requests', self.__dbconn, if_exists='append', index=False)
            return df_advance_requests['id'].to_list()

        return []

    def extract_reports(self, updated_at: List[str] = None, exported: bool = None) -> List[str]:
        """
        Extract reports from Fyle
        :param updated_at: Date string in yyyy-MM-ddTHH:mm:ss.SSSZ format along with operator in RHS colon pattern.
        :param exported: True for exported reports and False for unexported reports
        :return: List of report ids
        """
        self.logger.info('Extracting reports from Fyle.')

        reports = self.__connection.Reports.get_all(
            updated_at=updated_at,
            exported=exported
        )

        self.logger.info('%s reports extracted.', str(len(reports)))

        if reports:
            df_reports = pd.DataFrame(reports)

            df_reports['approved_by'] = df_reports['approved_by'].map(lambda report: report[0] if report else None)

            df_reports = df_reports[[
                'id', 'employee_id', 'employee_email', 'employee_code', 'state',
                'amount', 'purpose', 'claim_number', 'created_at', 'updated_at', 'approved_at',
                'reimbursed_at', 'trip_request_id', 'settlement_id', 'org_id', 'org_name', "verified",
                "exported", "approved_by", "created_by", "settled_at"
            ]]

            df_reports.to_sql('fyle_extract_reports', self.__dbconn, if_exists='append', index=False)
            return df_reports['id'].to_list()

        return []



import logging
import os
import sqlite3

import pytest

from test.common.utilities import get_mock_fyle, fyle_connect

from fyle_db_connector import FyleExtractConnector

logger = logging.getLogger(__name__)


@pytest.fixture
def mock_fyle():
    """
    make mock_fyle connection
    :return: mock_fyle
    """
    return get_mock_fyle()


@pytest.fixture(scope='module')
def fyle():
    """
    make fyle_connection
    :return: fyle_connection
    """
    return fyle_connect()


@pytest.fixture
def dbconn():
    """
    connect to a database
    :return: database connection
    """
    SQLITE_DB_FILE = '/tmp/test_fyle.db'
    if os.path.exists(SQLITE_DB_FILE):
        os.remove(SQLITE_DB_FILE)
    return sqlite3.connect(SQLITE_DB_FILE, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)


@pytest.fixture
def fec(fyle, dbconn):
    """
    make FyleExtractConnection
    :param fyle: fyle_connection
    :param dbconn: database_connection
    :return: FyleExtractConnector response
    """
    res = FyleExtractConnector(fyle_sdk_connection=fyle, dbconn=dbconn)
    res.create_tables()
    return res



import logging

from test.common.utilities import (dbconn_table_num_rows, dbconn_table_row_dict, dict_compare_keys, get_mock_fyle_empty)
from fyle_db_connector.extract import FyleExtractConnector

logger = logging.getLogger(__name__)


def test_employees(fec):
    """
    Test Extract Employees
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    ids = fec.extract_employees()
    fyle_data = fyle.Employees.get_all()
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_employees')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_employees') == len(fyle_data), 'row count mismatch'
    assert len(ids) == 3, 'return value messed up'


def test_settlements(fec):
    """
    Test Extract Settlements
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    ids = fec.extract_settlements()
    fyle_data = fyle.Settlements.get_all()
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_settlements')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_settlements') == len(fyle_data), 'row count mismatch'
    assert len(ids) == 5, 'return value messed up'


def test_reimbursements(fec):
    """
    Test Extract Reimbursements
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    ids = fec.extract_reimbursements()
    fyle_data = fyle.Reimbursements.get_all()
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_reimbursements')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_reimbursements') == len(fyle_data), 'row count mismatch'
    assert len(ids) == 3, 'return value messed up'


def test_expenses(fec):
    """
    Test Extract Expenses
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    ids = fec.extract_expenses(extract_custom_fields=False)
    fyle_data = fyle.Expenses.get_all()
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_expenses')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_expenses') == len(fyle_data), 'row count mismatch'
    assert len(ids) == 3, 'return value messed up'


def test_corporate_credit_card_expenses(fec):
    """
    Test Extract CorporateCreditCardExpenses
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    ids = fec.extract_corporate_credit_card_expenses()
    fyle_data = fyle.CorporateCreditCardExpenses.get_all()
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_corporate_credit_card_expenses')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_corporate_credit_card_expenses') == len(
        fyle_data), 'row count mismatch'
    assert len(ids) == 3, 'return value messed up'


def test_advances(fec):
    """
    Test Extract Advances
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    ids = fec.extract_advances()
    fyle_data = fyle.Advances.get_all()
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_advances')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_advances') == len(
        fyle_data), 'row count mismatch'
    assert len(ids) == 4, 'return value messed up'


def test_advance_requests(fec):
    """
    Test Extract AdvanceRequests
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    ids = fec.extract_advance_requests()
    fyle_data = fyle.AdvanceRequests.get_all()
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_advance_requests')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_advance_requests') == len(
        fyle_data), 'row count mismatch'
    assert len(ids) == 4, 'return value messed up'


def test_categories(fec):
    """
    Test Extract Categories
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    ids = fec.extract_categories()
    fyle_data = fyle.Categories.get()['data']
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_categories')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_categories') == len(fyle_data), 'row count mismatch'
    assert len(ids) == 26, 'return value messed up'


def test_projects(fec):
    """
    Test Extract Projects
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    ids = fec.extract_projects()
    fyle_data = fyle.Projects.get()['data']
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_projects')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_projects') == len(fyle_data), 'row count mismatch'
    assert len(ids) == 4, 'return value messed up'


def test_cost_centers(fec):
    """
    Test Extract CostCenters
    :param fec: fyle extract connection
    :return: None
    """
    dbconn = fec._FyleExtractConnector__dbconn
    fyle = fec._FyleExtractConnector__connection
    fyle_extract = FyleExtractConnector(fyle_sdk_connection=fyle, dbconn=dbconn)
    ids = fyle_extract.extract_cost_centers()
    fyle_data = fyle.CostCenters.get()['data']
    db_data = dbconn_table_row_dict(dbconn, 'fyle_extract_cost_centers')
    assert dict_compare_keys(db_data, fyle_data[0]) == [], 'db table has some columns that fyle doesnt'
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_cost_centers') == len(fyle_data), 'row count mismatch'
    assert len(ids) == 11, 'return value messed up'


def test_empty(dbconn):
    """
    Test Extract Empty
    :param dbconn: sqlite db connection
    :return: None
    """
    fyle = get_mock_fyle_empty()
    res = FyleExtractConnector(fyle_sdk_connection=fyle, dbconn=dbconn)
    res.create_tables()
    assert res.extract_expenses() == []



import logging
import os
import sqlite3
from os import path

import pytest

from test.common.utilities import get_mock_fyle
from fyle_db_connector.extract import FyleExtractConnector
from fyle_db_connector.load import FyleLoadConnector

logger = logging.getLogger(__name__)


@pytest.fixture
def fyle():
    return get_mock_fyle()


@pytest.fixture
def dbconn():
    SQLITE_DB_FILE = '/tmp/test_fyle.db'
    if os.path.exists(SQLITE_DB_FILE):
        os.remove(SQLITE_DB_FILE)
    conn = sqlite3.connect(SQLITE_DB_FILE, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
    return conn


@pytest.fixture
def fec(fyle, dbconn):
    res = FyleExtractConnector(fyle_sdk_connection=fyle, dbconn=dbconn)
    res.create_tables()
    return res


@pytest.fixture
def flc(fyle, dbconn):
    res = FyleLoadConnector(fyle_sdk_connection=fyle, dbconn=dbconn)
    res.create_tables()
    basepath = path.dirname(__file__)
    sqlpath = path.join(basepath, '../common/mock_db_load.sql')
    sql = open(sqlpath, 'r').read()
    dbconn.executescript(sql)
    return res



import logging
from test.common.utilities import (dict_compare_keys, get_mock_fyle_empty)
from fyle_db_connector.extract import FyleExtractConnector

logger = logging.getLogger(__name__)


def test_employees(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract employees
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.Employees.get_all()
    mock_data = mock_fyle.Employees.get_all()

    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_settlements(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract settlements
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.Settlements.get_all()
    mock_data = mock_fyle.Settlements.get_all()

    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_reimbursements(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract reimbursements
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.Reimbursements.get_all()
    mock_data = mock_fyle.Reimbursements.get_all()

    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_expenses(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract expenses
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.Expenses.get_all()
    mock_data = mock_fyle.Expenses.get_all()

    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_advances(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract advances
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.Advances.get_all()
    mock_data = mock_fyle.Advances.get_all()
    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_advance_requests(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract advance_requests
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.AdvanceRequests.get_all()
    mock_data = mock_fyle.AdvanceRequests.get_all()
    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_projects(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract projects
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.Projects.get()['data']
    mock_data = mock_fyle.Projects.get()['data']
    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_cost_centers(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract cost_centers
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.CostCenters.get()['data']
    mock_data = mock_fyle.CostCenters.get()['data']
    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_categories(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract categories
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.Categories.get()['data']
    mock_data = mock_fyle.Categories.get()['data']
    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_corporate_card_expenses(fyle, mock_fyle):
    """
    Test fyle_db_connector Extract corporate_credit_card_expenses
    :param fyle: fyle_extract instance
    :param mock_fyle: mock instance
    :return: None
    """
    data = fyle.CorporateCreditCardExpenses.get_all()
    mock_data = mock_fyle.CorporateCreditCardExpenses.get_all()
    assert dict_compare_keys(data[0], mock_data[0]) == [], 'real fyle has stuff that mock_fyle doesnt'
    assert dict_compare_keys(mock_data[0], data[0]) == [], 'mock_fyle has stuff that real fyle doesnt'


def test_empty(dbconn):
    """
    Test fyle_db_connector Extract empty_test
    :param dbconn: database connection
    :return: None
    """
    fyle = get_mock_fyle_empty()
    res = FyleExtractConnector(fyle_sdk_connection=fyle, dbconn=dbconn)
    res.create_tables()
    assert res.extract_expenses() == []



import pytest

import sqlite3
import logging
from test.common.utilities import dict_compare_keys, dbconn_table_num_rows

logger = logging.getLogger(__name__)


def test_fyle_mock_setup(fyle):
    """
    Testing mock data
    """
    logger.info('Testing mock data')

    assert fyle.Expenses.get_all()[0]['id'] == 'txNx96Gutp1X', 'fyle mock setup is broken'


def test_dbconn_mock_setup(dbconn):
    """
    Testing mock dbconn
    """
    logger.info('Testing mock dbconn')

    with pytest.raises(sqlite3.OperationalError) as e:
        rows = dbconn_table_num_rows(dbconn, 'fyle_extract_employees')


def test_fec_mock_setup(fec):
    """
    Testing Extract connector with mock instance
    """
    logger.info('Testing Extract connector with mock instance')

    # python magic to access private variable for testing db state
    dbconn = fec._FyleExtractConnector__dbconn
    assert dbconn_table_num_rows(dbconn, 'fyle_extract_employees') == 0, 'Unclean db'


def test_flc_mock_setup(flc, dbconn):
    """
    Testing Load Connector with mock instance
    """
    logger.info('Testing Load Connector with mock instance')

    flc.create_tables()
    sqlpath = './test/common/mock_db_load.sql'
    sql = open(sqlpath, 'r').read()
    dbconn.executescript(sql)

    assert dbconn_table_num_rows(dbconn, 'fyle_load_tpa_export_batches') == 2, 'Unclean db'
    assert dbconn_table_num_rows(dbconn, 'fyle_load_tpa_export_batch_lineitems') == 4, 'Unclean db'


def test_dict_compare():
    """
    Testing dict compare function
    """
    logger.info('Testing dict compare function')

    d1 = {
        'k1': 'xxx', 'k2': 2, 'k3': [1, 2], 'k4': {'k41': [2], 'k42': {'k421': 20}}
    }
    d2 = {
        'k1': 'xyx', 'k3': [1, 2], 'k4': {'k42': {'k421': 20}}
    }
    d3 = {
        'k1': 'xyz', 'k3': [3, 2], 'k4': {'k42': {'k421': 40}}
    }
    assert dict_compare_keys(d1, d2) == ['->k2', '->k4->k41'], 'not identifying diff properly'
    assert dict_compare_keys(d2, d3) == [], 'should return no diff'



"""
Load Unit Tests
"""
import logging

from test.common.utilities import dbconn_get_load_object_by_id, dbconn_get_load_object_by_batch_id

logger = logging.getLogger(__name__)


def test_fyle_load_tpa_export_batches(flc, dbconn):
    """
    tpa_export_batches load Unit Test
    """
    assert dbconn_get_load_object_by_id(dbconn, 'fyle_load_tpa_export_batches',
                                        'exbtCGWmKieHip'), 'fyle load_tpa_export id should not exist'
    sql = open('./test/common/mock_db_load.sql').read()
    dbconn.executescript(sql)

    response = flc.load_tpa_exports()
    assert response == ['exbtCGWmKieHip', 'exbt0EdBt3ojED', 'exbtCGWmKieHip', 'exbt0EdBt3ojED'], 'fyle load_tpa_export id not matching'


def test_fyle_load_tpa_export_batch_lineitems(flc, dbconn):
    """
    tpa_export_batch_lineitems load Unit Test
    """
    assert dbconn_get_load_object_by_batch_id(dbconn, 'fyle_load_tpa_export_batch_lineitems',
                                              'exbtCGWmKieHip'), 'fyle load_tpa_export_batch_lineitems  id should not exist'

    sql = open('./test/common/mock_db_load.sql').read()
    dbconn.executescript(sql)

    response = flc.load_tpa_export_batch(batch_id='exbtCGWmKieHip')
    assert response == ['exbtCGWmKieHip'], 'fyle load_tpa_export_batch_lineitems not matching'



"""
Build the PyPi package.
"""

import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='fyle-db-connector',
    version='0.13.0',
    author='Shwetabh Kumar',
    author_email='shwetabh.kumar@fyle.in',
    description='Connects Fyle to a database connector to transfer information to and fro.',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords=['fyle', 'db', 'python', 'sdk', 'sqlite', 'postgres'],
    url='https://github.com/fylein/fyle-db-connector',
    packages=setuptools.find_packages(),
    install_requires=[
        'pandas==0.25.2',
        'fylesdk>=0.13.0'
    ],
    classifiers=[
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ]
)



"""
FyleLoadConnector(): Connection between Fyle and Database
"""

import logging
from os import path
from typing import BinaryIO, List
import pandas as pd


class FyleLoadConnector:
    """
    - Extract Data from Database and load to Fyle
    """
    def __init__(self, fyle_sdk_connection, dbconn):
        self.__dbconn = dbconn
        self.__connection = fyle_sdk_connection
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info('Fyle connection established.')

    def create_tables(self):
        """
        Creates DB tables
        :return: None
        """
        basepath = path.dirname(__file__)
        ddl_path = path.join(basepath, 'load_ddl.sql')
        ddl_sql = open(ddl_path, 'r').read()
        self.__dbconn.executescript(ddl_sql)

    def __load_excel(self, file_path: str) -> str:
        """
        Upload Excel File to Fyle
        :param file_path: Absolute path for the excel file
        :return: returns file id
        """
        self.logger.info('Uploading excel to Fyle.')

        file_data = open(file_path, 'rb')

        file_path_tokenize = file_path.split('/')
        file_name = file_path_tokenize[len(file_path_tokenize) - 1]

        file_id = self.load_file(
            file_name,
            file_data,
            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        )

        self.logger.info('Excel file uploaded successfully.')
        return file_id

    def load_tpa_export_batch(self, batch_id, file_path: str = None, file_id: str = None) -> List:
        """
        Load a TPA Export Batch in Fyle
        :param batch_id: Batch ID
        :param file_path: Path of the export file
        :param file_id: Id of file already uploaded to Fyle
        :return: None
        """
        batches_df = pd.read_sql_query(sql=f"select * from fyle_load_tpa_export_batches where "
                                           f"id = '{batch_id}' limit 1", con=self.__dbconn)
        if len(batches_df) == 0:
            self.logger.info('No such batch')
            return []

        if not file_id:
            if file_path:
                file_id = self.__load_excel(file_path)
        batches_df['file_id'] = file_id

        batches_df['success'] = True
        batches = batches_df.to_dict(orient='records')
        batch = batches[0]
        self.logger.info('Uploading batch to Fyle')

        lineitems_df = pd.read_sql_query(
            sql=f"select * from fyle_load_tpa_export_batch_lineitems where batch_id = '{batch_id}'",
            con=self.__dbconn
        )

        if len(lineitems_df) == 0:
            self.logger.info('Batch %s has no lineitems, skipping', batch_id)
            return []

        lineitems = lineitems_df.to_dict(orient='records')

        batch_id = self.__connection.Exports.post_batch(batch)['id']
        self.logger.info('Batch successfully upload. Uploading Line items.')
        self.__connection.Exports.post_batch_lineitems(batch_id, lineitems)
        self.logger.info('%s Lineitems successfully uploaded.', len(lineitems))
        return batches_df['id'].to_list()

    def load_tpa_exports(self, file_path: str = None, file_id: str = None) -> List:
        """
        Load TPA Export Batches in Fyle
        :param file_path: Path of the export file
        :param file_id: Id of file already uploaded to Fyle
        :return: None
        """
        self.logger.warning('method deprecated - please use load_tpa_export_batch')

        batches_df = pd.read_sql_query(sql='select id from fyle_load_tpa_export_batches', con=self.__dbconn)
        if len(batches_df) == 0:
            self.logger.info('nothing to export')
            return []
        batches = batches_df.to_dict(orient='records')
        self.logger.info('Pushing %d batches to Fyle', len(batches))
        for batch in batches:
            self.load_tpa_export_batch(batch_id=batch['id'], file_path=file_path, file_id=file_id)
        return batches_df['id'].to_list()

    def load_file(self, file_name: str, file_data: BinaryIO, content_type: str) -> str:
        """
        Upload File to Fyle
        :param file_name: Name of the file
        :param file_data: Data of the fyle in bytes
        :param content_type: content type
        :return:
        """
        file_obj = self.__connection.Files.post(file_name)
        upload_url = self.__connection.Files.create_upload_url(file_obj['id'])['url']
        self.__connection.Files.upload_file_to_aws(content_type, file_data, upload_url)
        return file_obj['id']
